from django.http import JsonResponse
from django.shortcuts import render
from django.utils.translation import gettext as _
from social_core import exceptions as social_exceptions

from ..admin.views.errorpages import admin_csrf_failure, admin_error_page
from ..socialauth.providers import providers
from .exceptions import SocialAuthBanned, SocialAuthFailed
from .utils import get_exception_message, is_request_to_misago


def _is_ajax(request):
    # Request is AJAX when done with XMLHttpRequest
    if request.headers.get("x-requested-with") == "XMLHttpRequest":
        return True

    # Assume that request is AJAX when it has JSON payload
    return "application/json" in request.headers.get("content-type", "")


def _ajax_error(code, exception=None, default_message=None):
    return JsonResponse(
        {"detail": get_exception_message(exception, default_message)}, status=code
    )


@admin_error_page
def _error_page(request, code, exception=None, default_message=None):
    request.frontend_context.update({"CURRENT_LINK": "misago:error-%s" % code})

    return render(
        request,
        "misago/errorpages/%s.html" % code,
        {"message": get_exception_message(exception, default_message)},
        status=code,
    )


def banned(request, exception):
    ban = exception.ban

    request.frontend_context.update(
        {"MESSAGE": ban.get_serialized_message(), "CURRENT_LINK": "misago:error-banned"}
    )

    return render(request, "misago/errorpages/banned.html", {"ban": ban}, status=403)


def permission_denied(request, exception):
    if _is_ajax(request):
        return _ajax_error(403, exception, _("Permission denied."))
    return _error_page(request, 403, exception)


def page_not_found(request, exception):
    if _is_ajax(request):
        return _ajax_error(404, exception, "Not found.")
    return _error_page(request, 404, exception)


def social_auth_failed(request, exception):
    backend_name = None
    ban = None
    help_text = None
    message = None

    try:
        backend_name = exception.backend_name
    except AttributeError:
        pass

    try:
        exception_backend = exception.backend
        backend_name = providers.get_name(exception_backend.name)
    except AttributeError:
        pass

    if isinstance(exception, social_exceptions.NotAllowedToDisconnect):
        message = _(
            "A problem was encountered when disconnecting your account "
            "from the remote site."
        )
        help_text = _(
            "You are not allowed to disconnect your account from the other site, "
            "because currently it's the only way to sign in to your account."
        )
    elif backend_name:
        message = _(
            "A problem was encountered when signing you in using %(backend)s."
        ) % {"backend": backend_name}

        if isinstance(exception, social_exceptions.AuthCanceled):
            help_text = _("The sign in process has been canceled by user.")
        if isinstance(exception, social_exceptions.AuthUnreachableProvider):
            help_text = _("The other service could not be reached.")
        if isinstance(exception, SocialAuthFailed):
            help_text = exception.message
        if isinstance(exception, SocialAuthBanned):
            ban = exception.ban
    else:
        message = _("Unexpected problem has been encountered during sign in process.")

    return render(
        request,
        "misago/errorpages/social.html",
        {
            "backend_name": backend_name,
            "ban": ban,
            "message": message,
            "help_text": help_text,
        },
        status=403,
    )


@admin_csrf_failure
def csrf_failure(request, reason=""):
    if _is_ajax(request):
        return JsonResponse(
            {
                "detail": _(
                    "Your request was rejected because your browser didn't "
                    "send the CSRF cookie, or the cookie sent was invalid."
                )
            },
            status=403,
        )

    return render(request, "misago/errorpages/csrf_failure.html", status=403)


def not_allowed(request):
    return _error_page(request, 405)


# Decorators for custom error page handlers
def shared_403_exception_handler(f):
    def page_decorator(request, *args, **kwargs):
        if is_request_to_misago(request):
            return permission_denied(request, *args, **kwargs)
        return f(request, *args, **kwargs)

    return page_decorator


def shared_404_exception_handler(f):
    def page_decorator(request, *args, **kwargs):
        if is_request_to_misago(request):
            return page_not_found(request, *args, **kwargs)
        return f(request, *args, **kwargs)

    return page_decorator



# Generated by Django 1.11.11 on 2018-03-31 22:08
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("misago_users", "0010_user_profile_fields")]

    operations = [
        migrations.AddField(
            model_name="user",
            name="is_deleting_account",
            field=models.BooleanField(default=False),
        )
    ]



from django.core.exceptions import PermissionDenied
from django.test import TestCase
from django.test.client import RequestFactory
from django.urls import reverse

from ..utils import (
    clean_ids_list,
    clean_return_path,
    format_plaintext_for_html,
    get_exception_message,
    get_host_from_address,
    is_referer_local,
    is_request_to_misago,
    parse_iso8601_string,
    slugify,
)


class IsRequestToMisagoTests(TestCase):
    def test_is_request_to_misago(self):
        """
        is_request_to_misago correctly detects requests directed at Misago
        """
        VALID_PATHS = ("/", "/threads/")
        INVALID_PATHS = ("", "somewhere/")

        misago_prefix = reverse("misago:index")

        for path in VALID_PATHS:
            request = RequestFactory().get("/")
            request.path = path
            self.assertTrue(
                is_request_to_misago(request),
                '"%s" is not overlapped by "%s"' % (path, misago_prefix),
            )

        for path in INVALID_PATHS:
            request = RequestFactory().get("/")
            request.path = path
            self.assertFalse(
                is_request_to_misago(request),
                '"%s" is overlapped by "%s"' % (path, misago_prefix),
            )


class SlugifyTests(TestCase):
    def test_valid_slugify_output(self):
        """Misago's slugify correctly slugifies string"""
        test_cases = [
            ("User", "user"),
            ("Eric The Fish", "eric-the-fish"),
            ("John   Snow", "john-snow"),
            ("J0n", "j0n"),
            ("An###ne", "anne"),
            ("S**t", "st"),
            ("Łók", "lok"),
        ]

        for original, slug in test_cases:
            self.assertEqual(slugify(original), slug)


class ParseIso8601StringTests(TestCase):
    def test_valid_input(self):
        """util parses iso 8601 strings"""
        INPUTS = [
            "2016-10-22T20:55:39.185085Z",
            "2016-10-22T20:55:39.185085-01:00",
            "2016-10-22T20:55:39-01:00",
            "2016-10-22T20:55:39.185085+01:00",
        ]

        for test_input in INPUTS:
            self.assertTrue(parse_iso8601_string(test_input))

    def test_invalid_input(self):
        """util throws ValueError on invalid input"""
        INPUTS = [
            "",
            "2016-10-22",
            "2016-10-22T30:55:39.185085+11:00",
            "2016-10-22T20:55:39.18SSSSS5085Z",
        ]

        for test_input in INPUTS:
            with self.assertRaises(ValueError):
                self.assertTrue(parse_iso8601_string(test_input))


PLAINTEXT_FORMAT_CASES = [
    ("Lorem ipsum.", "<p>Lorem ipsum.</p>"),
    ("Lorem <b>ipsum</b>.", "<p>Lorem &lt;b&gt;ipsum&lt;/b&gt;.</p>"),
    ('Lorem "ipsum" dolor met.', "<p>Lorem &quot;ipsum&quot; dolor met.</p>"),
    ("Lorem ipsum.\nDolor met.", "<p>Lorem ipsum.<br>Dolor met.</p>"),
    ("Lorem ipsum.\n\nDolor met.", "<p>Lorem ipsum.</p>\n\n<p>Dolor met.</p>"),
    (
        "http://misago-project.org/login/",
        (
            "<p>"
            '<a href="http://misago-project.org/login/">'
            "http://misago-project.org/login/"
            "</a>"
            "</p>"
        ),
    ),
]


class FormatPlaintextForHtmlTests(TestCase):
    def test_format_plaintext_for_html(self):
        """format_plaintext_for_html correctly formats plaintext for html"""
        for plaintext, html in PLAINTEXT_FORMAT_CASES:
            output = format_plaintext_for_html(plaintext)

            assertion_message = """
format_plaintext_for_html failed to produce expected output:

expected:   %s
return:     %s
""" % (
                html,
                output,
            )
            self.assertEqual(output, html, assertion_message)


class MockRequest:
    scheme = "http"

    def __init__(self, method, meta=None, post=None):
        self.method = method
        self.META = meta or {}
        self.POST = post or {}


class CleanReturnPathTests(TestCase):
    def test_get_request(self):
        """clean_return_path works for GET requests"""
        bad_request = MockRequest(
            "GET",
            {"HTTP_REFERER": "http://cookies.com", "HTTP_HOST": "misago-project.org"},
        )
        self.assertIsNone(clean_return_path(bad_request))

        bad_request = MockRequest(
            "GET",
            {
                "HTTP_REFERER": "https://misago-project.org/",
                "HTTP_HOST": "misago-project.org/",
            },
        )
        self.assertIsNone(clean_return_path(bad_request))

        bad_request = MockRequest(
            "GET",
            {
                "HTTP_REFERER": "https://misago-project.org/",
                "HTTP_HOST": "misago-project.org/assadsa/",
            },
        )
        self.assertIsNone(clean_return_path(bad_request))

        ok_request = MockRequest(
            "GET",
            {
                "HTTP_REFERER": "http://misago-project.org/",
                "HTTP_HOST": "misago-project.org/",
            },
        )
        self.assertEqual(clean_return_path(ok_request), "/")

        ok_request = MockRequest(
            "GET",
            {
                "HTTP_REFERER": "http://misago-project.org/login/",
                "HTTP_HOST": "misago-project.org/",
            },
        )
        self.assertEqual(clean_return_path(ok_request), "/login/")

    def test_post_request(self):
        """clean_return_path works for POST requests"""
        bad_request = MockRequest(
            "POST",
            {
                "HTTP_REFERER": "http://misago-project.org/",
                "HTTP_HOST": "misago-project.org/",
            },
            {"return_path": "/sdasdsa/"},
        )
        self.assertIsNone(clean_return_path(bad_request))

        ok_request = MockRequest(
            "POST",
            {
                "HTTP_REFERER": "http://misago-project.org/",
                "HTTP_HOST": "misago-project.org/",
            },
            {"return_path": "/login/"},
        )
        self.assertEqual(clean_return_path(ok_request), "/login/")


class IsRefererLocalTests(TestCase):
    def test_local_referers(self):
        """local referers return true"""
        ok_request = MockRequest(
            "GET",
            {
                "HTTP_REFERER": "http://misago-project.org/",
                "HTTP_HOST": "misago-project.org/",
            },
        )
        self.assertTrue(is_referer_local(ok_request))

        ok_request = MockRequest(
            "GET",
            {
                "HTTP_REFERER": "http://misago-project.org/",
                "HTTP_HOST": "misago-project.org/",
            },
        )
        self.assertTrue(is_referer_local(ok_request))

        ok_request = MockRequest(
            "GET",
            {
                "HTTP_REFERER": "http://misago-project.org/login/",
                "HTTP_HOST": "misago-project.org/",
            },
        )
        self.assertTrue(is_referer_local(ok_request))

    def test_foreign_referers(self):
        """non-local referers return false"""
        bad_request = MockRequest(
            "GET",
            {
                "HTTP_REFERER": "http://else-project.org/",
                "HTTP_HOST": "misago-project.org/",
            },
        )
        self.assertFalse(is_referer_local(bad_request))

        bad_request = MockRequest(
            "GET",
            {
                "HTTP_REFERER": "https://misago-project.org/",
                "HTTP_HOST": "misago-project.org/",
            },
        )
        self.assertFalse(is_referer_local(bad_request))

        bad_request = MockRequest(
            "GET",
            {
                "HTTP_REFERER": "http://misago-project.org/",
                "HTTP_HOST": "misago-project.org/assadsa/",
            },
        )
        self.assertFalse(is_referer_local(bad_request))


class GetExceptionMessageTests(TestCase):
    def test_no_args(self):
        """both of helper args are optional"""
        message = get_exception_message()
        self.assertIsNone(message)

    def test_no_default_message(self):
        """helper's default message arg is optional"""
        message = get_exception_message(PermissionDenied("Lorem Ipsum"))
        self.assertEqual(message, "Lorem Ipsum")

        message = get_exception_message(PermissionDenied())
        self.assertIsNone(message)

    def test_default_message(self):
        """helper's default message arg is used"""
        message = get_exception_message(PermissionDenied("Lorem Ipsum"), "Default")
        self.assertEqual(message, "Lorem Ipsum")

        message = get_exception_message(PermissionDenied(), "Default")
        self.assertEqual(message, "Default")

        message = get_exception_message(default_message="Lorem Ipsum")
        self.assertEqual(message, "Lorem Ipsum")


class CleanIdsListTests(TestCase):
    def test_valid_list(self):
        """list of valid ids is cleaned"""
        self.assertEqual(clean_ids_list(["1", 3, "42"], None), [1, 3, 42])

    def test_empty_list(self):
        """empty list passes validation"""
        self.assertEqual(clean_ids_list([], None), [])

    def test_string_list(self):
        """string list passes validation"""
        self.assertEqual(clean_ids_list("1234", None), [1, 2, 3, 4])

    def test_message(self):
        """utility uses passed message for exception"""
        with self.assertRaisesMessage(PermissionDenied, "Test error message!"):
            clean_ids_list(None, "Test error message!")

    def test_invalid_inputs(self):
        """utility raises exception for invalid inputs"""
        INVALID_INPUTS = (
            None,
            "abc",
            [None],
            [1, 2, "a", 4],
            [1, None, 3],
            {1: 2, "a": 4},
        )

        for invalid_input in INVALID_INPUTS:
            with self.assertRaisesMessage(PermissionDenied, "Test error message!"):
                clean_ids_list(invalid_input, "Test error message!")


class GetHostFromAddressTests(TestCase):
    def test_none(self):
        """get_host_from_address returns None for None"""
        result = get_host_from_address(None)
        self.assertIsNone(result)

    def test_empty_string(self):
        """get_host_from_address returns None for empty string"""
        result = get_host_from_address("")
        self.assertIsNone(result)

    def test_hostname(self):
        """get_host_from_address returns hostname unchanged"""
        result = get_host_from_address("hostname")
        self.assertEqual(result, "hostname")

    def test_hostname_with_trailing_slash(self):
        """get_host_from_address returns hostname for hostname with trailing slash"""
        result = get_host_from_address("//hostname")
        self.assertEqual(result, "hostname")

    def test_hostname_with_port(self):
        """get_host_from_address returns hostname for hostname with port"""
        result = get_host_from_address("hostname:8888")
        self.assertEqual(result, "hostname")

    def test_hostname_with_port_path_and_protocol(self):
        """get_host_from_address returns hostname for hostname with port and path"""
        result = get_host_from_address("https://hostname:8888/somewhere/else/")
        self.assertEqual(result, "hostname")



# Generated by Django 3.2.15 on 2023-01-04 12:15

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Subject",
            fields=[
                (
                    "sub",
                    models.CharField(max_length=36, primary_key=True, serialize=False),
                ),
                ("created_on", models.DateTimeField(default=django.utils.timezone.now)),
                (
                    "last_used_on",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
    ]



from rest_framework import serializers

from .errorpages import not_allowed


def require_POST(f):
    def decorator(request, *args, **kwargs):
        if not request.method == "POST":
            return not_allowed(request)
        return f(request, *args, **kwargs)

    return decorator


def require_dict_data(f):
    def decorator(request, *args, **kwargs):
        if request.method == "POST":
            DummySerializer(data=request.data).is_valid(raise_exception=True)
        return f(request, *args, **kwargs)

    return decorator


class DummySerializer(serializers.Serializer):
    pass



from rest_framework.response import Response

from ...moderation import posts as moderation
from ...permissions import (
    allow_delete_best_answer,
    allow_delete_event,
    allow_delete_post,
)
from ...serializers import DeletePostsSerializer


def delete_post(request, thread, post):
    if post.is_event:
        allow_delete_event(request.user_acl, post)
    else:
        allow_delete_best_answer(request.user_acl, post)
        allow_delete_post(request.user_acl, post)

    moderation.delete_post(request.user, post)

    sync_related(thread)
    return Response({})


def delete_bulk(request, thread):
    serializer = DeletePostsSerializer(
        data={"posts": request.data},
        context={
            "settings": request.settings,
            "thread": thread,
            "user_acl": request.user_acl,
        },
    )

    if not serializer.is_valid():
        if "posts" in serializer.errors:
            errors = serializer.errors["posts"]
        else:
            errors = list(serializer.errors.values())[0]
        # Fix for KeyError - errors[0]
        try:
            errors = errors[0]
        except KeyError:
            if errors and isinstance(errors, dict):
                errors = list(errors.values())[0][0]
        return Response({"detail": errors}, status=400)

    for post in serializer.validated_data["posts"]:
        post.delete()

    sync_related(thread)

    return Response({})


def sync_related(thread):
    thread.synchronize()
    thread.save()

    thread.category.synchronize()
    thread.category.save()



from django.urls import reverse

from .. import test
from ...categories.models import Category
from ...readtracker import poststracker
from ..models import Poll, PollVote, Thread
from ..test import patch_category_acl, patch_other_category_acl
from .test_threads_api import ThreadsApiTestCase


class ThreadMergeApiTests(ThreadsApiTestCase):
    def setUp(self):
        super().setUp()

        Category(name="Other Category", slug="other-category").insert_at(
            self.category, position="last-child", save=True
        )
        self.other_category = Category.objects.get(slug="other-category")

        self.api_link = reverse(
            "misago:api:thread-merge", kwargs={"pk": self.thread.pk}
        )

    @patch_category_acl({"can_merge_threads": False})
    def test_merge_no_permission(self):
        """api validates if thread can be merged with other one"""
        response = self.client.post(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "You can't merge threads in this category."}
        )

    @patch_category_acl({"can_merge_threads": True})
    def test_merge_no_url(self):
        """api validates if thread url was given"""
        response = self.client.post(self.api_link)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"detail": "Enter link to new thread."})

    @patch_category_acl({"can_merge_threads": True})
    def test_invalid_url(self):
        """api validates thread url"""
        response = self.client.post(
            self.api_link, {"other_thread": self.user.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"detail": "This is not a valid thread link."}
        )

    @patch_category_acl({"can_merge_threads": True})
    def test_current_other_thread(self):
        """api validates if thread url given is to current thread"""
        response = self.client.post(
            self.api_link, {"other_thread": self.thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"detail": "You can't merge thread with itself."}
        )

    @patch_other_category_acl()
    @patch_category_acl({"can_merge_threads": True})
    def test_other_thread_exists(self):
        """api validates if other thread exists"""
        other_thread = test.post_thread(self.other_category)
        other_other_thread = other_thread.get_absolute_url()
        other_thread.delete()

        response = self.client.post(self.api_link, {"other_thread": other_other_thread})
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "detail": (
                    "The thread you have entered link to doesn't exist "
                    "or you don't have permission to see it."
                )
            },
        )

    @patch_other_category_acl({"can_see": False})
    @patch_category_acl({"can_merge_threads": True})
    def test_other_thread_is_invisible(self):
        """api validates if other thread is visible"""
        other_thread = test.post_thread(self.other_category)

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "detail": (
                    "The thread you have entered link to doesn't exist "
                    "or you don't have permission to see it."
                )
            },
        )

    @patch_other_category_acl({"can_merge_threads": False})
    @patch_category_acl({"can_merge_threads": True})
    def test_other_thread_isnt_mergeable(self):
        """api validates if other thread can be merged"""
        other_thread = test.post_thread(self.other_category)

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"detail": "Other thread can't be merged with."}
        )

    @patch_other_category_acl({"can_merge_threads": True, "can_close_threads": False})
    @patch_category_acl({"can_merge_threads": True})
    def test_thread_category_is_closed(self):
        """api validates if thread's category is open"""
        other_thread = test.post_thread(self.other_category)

        self.category.is_closed = True
        self.category.save()

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "This category is closed. You can't merge it's threads."},
        )

    @patch_other_category_acl({"can_merge_threads": True, "can_close_threads": False})
    @patch_category_acl({"can_merge_threads": True})
    def test_thread_is_closed(self):
        """api validates if thread is open"""
        other_thread = test.post_thread(self.other_category)

        self.thread.is_closed = True
        self.thread.save()

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "This thread is closed. You can't merge it with other threads."},
        )

    @patch_other_category_acl({"can_merge_threads": True, "can_close_threads": False})
    @patch_category_acl({"can_merge_threads": True})
    def test_other_thread_category_is_closed(self):
        """api validates if other thread's category is open"""
        other_thread = test.post_thread(self.other_category)

        self.other_category.is_closed = True
        self.other_category.save()

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "Other thread's category is closed. You can't merge with it."},
        )

    @patch_other_category_acl({"can_merge_threads": True, "can_close_threads": False})
    @patch_category_acl({"can_merge_threads": True})
    def test_other_thread_is_closed(self):
        """api validates if other thread is open"""
        other_thread = test.post_thread(self.other_category)

        other_thread.is_closed = True
        other_thread.save()

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "Other thread is closed and can't be merged with."},
        )

    @patch_other_category_acl({"can_merge_threads": True, "can_reply_threads": False})
    @patch_category_acl({"can_merge_threads": True})
    def test_other_thread_isnt_replyable(self):
        """api validates if other thread can be replied, which is condition for merge"""
        other_thread = test.post_thread(self.other_category)

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "You can't merge this thread into thread you can't reply."},
        )

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    def test_merge_threads(self):
        """api merges two threads successfully"""
        other_thread = test.post_thread(self.other_category)

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "id": other_thread.id,
                "title": other_thread.title,
                "url": other_thread.get_absolute_url(),
            },
        )

        # other thread has two posts and an event now
        self.assertEqual(other_thread.post_set.count(), 3)

        # first thread is gone
        with self.assertRaises(Thread.DoesNotExist):
            Thread.objects.get(pk=self.thread.pk)

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    def test_merge_threads_kept_reads(self):
        """api keeps both threads readtrackers after merge"""
        other_thread = test.post_thread(self.other_category)

        poststracker.save_read(self.user, self.thread.first_post)
        poststracker.save_read(self.user, other_thread.first_post)

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "id": other_thread.id,
                "title": other_thread.title,
                "url": other_thread.get_absolute_url(),
            },
        )

        # posts reads are kept
        postreads = self.user.postread_set.filter(post__is_event=False).order_by("id")

        self.assertEqual(
            list(postreads.values_list("post_id", flat=True)),
            [self.thread.first_post_id, other_thread.first_post_id],
        )
        self.assertEqual(postreads.filter(thread=other_thread).count(), 2)
        self.assertEqual(postreads.filter(category=self.other_category).count(), 2)

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    def test_merge_threads_kept_subs(self):
        """api keeps other thread's subscription after merge"""
        other_thread = test.post_thread(self.other_category)

        self.user.subscription_set.create(
            thread=self.thread,
            category=self.thread.category,
            last_read_on=self.thread.last_post_on,
            send_email=False,
        )

        self.assertEqual(self.user.subscription_set.count(), 1)
        self.user.subscription_set.get(thread=self.thread)
        self.user.subscription_set.get(category=self.category)

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "id": other_thread.id,
                "title": other_thread.title,
                "url": other_thread.get_absolute_url(),
            },
        )

        # subscriptions are kept
        self.assertEqual(self.user.subscription_set.count(), 1)
        self.user.subscription_set.get(thread=other_thread)
        self.user.subscription_set.get(category=self.other_category)

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    def test_merge_threads_moved_subs(self):
        """api keeps other thread's subscription after merge"""
        other_thread = test.post_thread(self.other_category)

        self.user.subscription_set.create(
            thread=other_thread,
            category=other_thread.category,
            last_read_on=other_thread.last_post_on,
            send_email=False,
        )

        self.assertEqual(self.user.subscription_set.count(), 1)
        self.user.subscription_set.get(thread=other_thread)
        self.user.subscription_set.get(category=self.other_category)

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "id": other_thread.id,
                "title": other_thread.title,
                "url": other_thread.get_absolute_url(),
            },
        )

        # subscriptions are kept
        self.assertEqual(self.user.subscription_set.count(), 1)
        self.user.subscription_set.get(thread=other_thread)
        self.user.subscription_set.get(category=self.other_category)

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    def test_merge_threads_handle_subs_colision(self):
        """api resolves conflicting thread subscriptions after merge"""
        self.user.subscription_set.create(
            thread=self.thread,
            category=self.thread.category,
            last_read_on=self.thread.last_post_on,
            send_email=False,
        )

        other_thread = test.post_thread(self.other_category)

        self.user.subscription_set.create(
            thread=other_thread,
            category=other_thread.category,
            last_read_on=other_thread.last_post_on,
            send_email=False,
        )

        self.assertEqual(self.user.subscription_set.count(), 2)
        self.user.subscription_set.get(thread=self.thread)
        self.user.subscription_set.get(category=self.category)
        self.user.subscription_set.get(thread=other_thread)
        self.user.subscription_set.get(category=self.other_category)

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "id": other_thread.id,
                "title": other_thread.title,
                "url": other_thread.get_absolute_url(),
            },
        )

        # subscriptions are kept
        self.assertEqual(self.user.subscription_set.count(), 1)
        self.user.subscription_set.get(thread=other_thread)
        self.user.subscription_set.get(category=self.other_category)

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    def test_merge_threads_kept_best_answer(self):
        """api merges two threads successfully, keeping best answer from old thread"""
        other_thread = test.post_thread(self.other_category)
        best_answer = test.reply_thread(other_thread)
        other_thread.set_best_answer(self.user, best_answer)
        other_thread.save()

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "id": other_thread.id,
                "title": other_thread.title,
                "url": other_thread.get_absolute_url(),
            },
        )

        # other thread has three posts and an event now
        self.assertEqual(other_thread.post_set.count(), 4)

        # first thread is gone
        with self.assertRaises(Thread.DoesNotExist):
            Thread.objects.get(pk=self.thread.pk)

        # best answer is kept in other thread
        other_thread = Thread.objects.get(pk=other_thread.pk)
        self.assertEqual(other_thread.best_answer, best_answer)

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    def test_merge_threads_moved_best_answer(self):
        """api merges two threads successfully, moving best answer to old thread"""
        other_thread = test.post_thread(self.other_category)

        best_answer = test.reply_thread(self.thread)
        self.thread.set_best_answer(self.user, best_answer)
        self.thread.save()

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "id": other_thread.id,
                "title": other_thread.title,
                "url": other_thread.get_absolute_url(),
            },
        )

        # other thread has three posts and an event now
        self.assertEqual(other_thread.post_set.count(), 4)

        # first thread is gone
        with self.assertRaises(Thread.DoesNotExist):
            Thread.objects.get(pk=self.thread.pk)

        # best answer is kept in other thread
        other_thread = Thread.objects.get(pk=other_thread.pk)
        self.assertEqual(other_thread.best_answer, best_answer)

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    def test_merge_threads_merge_conflict_best_answer(self):
        """api errors on merge conflict, returning list of available best answers"""
        best_answer = test.reply_thread(self.thread)
        self.thread.set_best_answer(self.user, best_answer)
        self.thread.save()

        other_thread = test.post_thread(self.other_category)
        other_best_answer = test.reply_thread(other_thread)
        other_thread.set_best_answer(self.user, other_best_answer)
        other_thread.save()

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "best_answers": [
                    ["0", "Unmark all best answers"],
                    [str(self.thread.id), self.thread.title],
                    [str(other_thread.id), other_thread.title],
                ]
            },
        )

        # best answers were untouched
        self.assertEqual(self.thread.post_set.count(), 2)
        self.assertEqual(other_thread.post_set.count(), 2)
        self.assertEqual(
            Thread.objects.get(pk=self.thread.pk).best_answer_id, best_answer.id
        )
        self.assertEqual(
            Thread.objects.get(pk=other_thread.pk).best_answer_id, other_best_answer.id
        )

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    def test_threads_merge_conflict_best_answer_invalid_resolution(self):
        """api errors on invalid merge conflict resolution"""
        best_answer = test.reply_thread(self.thread)
        self.thread.set_best_answer(self.user, best_answer)
        self.thread.save()

        other_thread = test.post_thread(self.other_category)
        other_best_answer = test.reply_thread(other_thread)
        other_thread.set_best_answer(self.user, other_best_answer)
        other_thread.save()

        response = self.client.post(
            self.api_link,
            {
                "other_thread": other_thread.get_absolute_url(),
                "best_answer": other_thread.id + 10,
            },
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"detail": "Invalid choice."})

        # best answers were untouched
        self.assertEqual(self.thread.post_set.count(), 2)
        self.assertEqual(other_thread.post_set.count(), 2)
        self.assertEqual(
            Thread.objects.get(pk=self.thread.pk).best_answer_id, best_answer.id
        )
        self.assertEqual(
            Thread.objects.get(pk=other_thread.pk).best_answer_id, other_best_answer.id
        )

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    def test_threads_merge_conflict_unmark_all_best_answers(self):
        """api unmarks all best answers when unmark all choice is selected"""
        best_answer = test.reply_thread(self.thread)
        self.thread.set_best_answer(self.user, best_answer)
        self.thread.save()

        other_thread = test.post_thread(self.other_category)
        other_best_answer = test.reply_thread(other_thread)
        other_thread.set_best_answer(self.user, other_best_answer)
        other_thread.save()

        response = self.client.post(
            self.api_link,
            {"other_thread": other_thread.get_absolute_url(), "best_answer": 0},
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "id": other_thread.id,
                "title": other_thread.title,
                "url": other_thread.get_absolute_url(),
            },
        )

        # other thread has four posts and an event now
        self.assertEqual(other_thread.post_set.count(), 5)

        # first thread is gone
        with self.assertRaises(Thread.DoesNotExist):
            Thread.objects.get(pk=self.thread.pk)

        # final thread has no marked best answer
        self.assertIsNone(Thread.objects.get(pk=other_thread.pk).best_answer_id)

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    def test_threads_merge_conflict_keep_first_best_answer(self):
        """api unmarks other best answer on merge"""
        best_answer = test.reply_thread(self.thread)
        self.thread.set_best_answer(self.user, best_answer)
        self.thread.save()

        other_thread = test.post_thread(self.other_category)
        other_best_answer = test.reply_thread(other_thread)
        other_thread.set_best_answer(self.user, other_best_answer)
        other_thread.save()

        response = self.client.post(
            self.api_link,
            {
                "other_thread": other_thread.get_absolute_url(),
                "best_answer": self.thread.pk,
            },
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "id": other_thread.id,
                "title": other_thread.title,
                "url": other_thread.get_absolute_url(),
            },
        )

        # other thread has four posts and an event now
        self.assertEqual(other_thread.post_set.count(), 5)

        # first thread is gone
        with self.assertRaises(Thread.DoesNotExist):
            Thread.objects.get(pk=self.thread.pk)

        # other thread's best answer was unchanged
        self.assertEqual(
            Thread.objects.get(pk=other_thread.pk).best_answer_id, best_answer.id
        )

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    def test_threads_merge_conflict_keep_other_best_answer(self):
        """api unmarks first best answer on merge"""
        best_answer = test.reply_thread(self.thread)
        self.thread.set_best_answer(self.user, best_answer)
        self.thread.save()

        other_thread = test.post_thread(self.other_category)
        other_best_answer = test.reply_thread(other_thread)
        other_thread.set_best_answer(self.user, other_best_answer)
        other_thread.save()

        response = self.client.post(
            self.api_link,
            {
                "other_thread": other_thread.get_absolute_url(),
                "best_answer": other_thread.pk,
            },
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "id": other_thread.id,
                "title": other_thread.title,
                "url": other_thread.get_absolute_url(),
            },
        )

        # other thread has four posts and an event now
        self.assertEqual(other_thread.post_set.count(), 5)

        # first thread is gone
        with self.assertRaises(Thread.DoesNotExist):
            Thread.objects.get(pk=self.thread.pk)

        # other thread's best answer was changed to merged in thread's answer
        self.assertEqual(
            Thread.objects.get(pk=other_thread.pk).best_answer_id, other_best_answer.id
        )

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    def test_merge_threads_kept_poll(self):
        """api merges two threads successfully, keeping poll from other thread"""
        other_thread = test.post_thread(self.other_category)
        poll = test.post_poll(other_thread, self.user)

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "id": other_thread.id,
                "title": other_thread.title,
                "url": other_thread.get_absolute_url(),
            },
        )

        # other thread has two posts and an event now
        self.assertEqual(other_thread.post_set.count(), 3)

        # first thread is gone
        with self.assertRaises(Thread.DoesNotExist):
            Thread.objects.get(pk=self.thread.pk)

        # poll and its votes were kept
        self.assertEqual(
            Poll.objects.filter(pk=poll.pk, thread=other_thread).count(), 1
        )
        self.assertEqual(
            PollVote.objects.filter(poll=poll, thread=other_thread).count(), 4
        )

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    def test_merge_threads_moved_poll(self):
        """api merges two threads successfully, moving poll from old thread"""
        other_thread = test.post_thread(self.other_category)
        poll = test.post_poll(self.thread, self.user)

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "id": other_thread.id,
                "title": other_thread.title,
                "url": other_thread.get_absolute_url(),
            },
        )

        # other thread has two posts and an event now
        self.assertEqual(other_thread.post_set.count(), 3)

        # first thread is gone
        with self.assertRaises(Thread.DoesNotExist):
            Thread.objects.get(pk=self.thread.pk)

        # poll and its votes were moved
        self.assertEqual(
            Poll.objects.filter(pk=poll.pk, thread=other_thread).count(), 1
        )
        self.assertEqual(
            PollVote.objects.filter(poll=poll, thread=other_thread).count(), 4
        )

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    def test_threads_merge_conflict_polls(self):
        """api errors on merge conflict, returning list of available polls"""
        other_thread = test.post_thread(self.other_category)
        poll = test.post_poll(self.thread, self.user)
        other_poll = test.post_poll(other_thread, self.user)

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "polls": [
                    ["0", "Delete all polls"],
                    [str(poll.pk), "%s (%s)" % (poll.question, poll.thread.title)],
                    [
                        str(other_poll.pk),
                        "%s (%s)" % (other_poll.question, other_poll.thread.title),
                    ],
                ]
            },
        )

        # polls and votes were untouched
        self.assertEqual(Poll.objects.count(), 2)
        self.assertEqual(PollVote.objects.count(), 8)

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    def test_threads_merge_conflict_poll_invalid_resolution(self):
        """api errors on invalid merge conflict resolution"""
        other_thread = test.post_thread(self.other_category)

        test.post_poll(self.thread, self.user)
        test.post_poll(other_thread, self.user)

        response = self.client.post(
            self.api_link,
            {
                "other_thread": other_thread.get_absolute_url(),
                "poll": Poll.objects.all()[0].pk + 10,
            },
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"detail": "Invalid choice."})

        # polls and votes were untouched
        self.assertEqual(Poll.objects.count(), 2)
        self.assertEqual(PollVote.objects.count(), 8)

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    def test_threads_merge_conflict_delete_all_polls(self):
        """api deletes all polls when delete all choice is selected"""
        other_thread = test.post_thread(self.other_category)
        test.post_poll(self.thread, self.user)
        test.post_poll(other_thread, self.user)

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url(), "poll": 0}
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "id": other_thread.id,
                "title": other_thread.title,
                "url": other_thread.get_absolute_url(),
            },
        )

        # other thread has two posts and an event now
        self.assertEqual(other_thread.post_set.count(), 3)

        # first thread is gone
        with self.assertRaises(Thread.DoesNotExist):
            Thread.objects.get(pk=self.thread.pk)

        # polls and votes are gone
        self.assertEqual(Poll.objects.count(), 0)
        self.assertEqual(PollVote.objects.count(), 0)

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    def test_threads_merge_conflict_keep_first_poll(self):
        """api deletes other poll on merge"""
        other_thread = test.post_thread(self.other_category)
        poll = test.post_poll(self.thread, self.user)
        other_poll = test.post_poll(other_thread, self.user)

        response = self.client.post(
            self.api_link,
            {"other_thread": other_thread.get_absolute_url(), "poll": poll.pk},
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "id": other_thread.id,
                "title": other_thread.title,
                "url": other_thread.get_absolute_url(),
            },
        )

        # other thread has two posts and an event now
        self.assertEqual(other_thread.post_set.count(), 3)

        # first thread is gone
        with self.assertRaises(Thread.DoesNotExist):
            Thread.objects.get(pk=self.thread.pk)

        # other poll and its votes are gone
        self.assertEqual(Poll.objects.filter(thread=self.thread).count(), 0)
        self.assertEqual(PollVote.objects.filter(thread=self.thread).count(), 0)

        self.assertEqual(Poll.objects.filter(thread=other_thread).count(), 1)
        self.assertEqual(PollVote.objects.filter(thread=other_thread).count(), 4)

        Poll.objects.get(pk=poll.pk)
        with self.assertRaises(Poll.DoesNotExist):
            Poll.objects.get(pk=other_poll.pk)

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    def test_threads_merge_conflict_keep_other_poll(self):
        """api deletes first poll on merge"""
        other_thread = test.post_thread(self.other_category)
        poll = test.post_poll(self.thread, self.user)
        other_poll = test.post_poll(other_thread, self.user)

        response = self.client.post(
            self.api_link,
            {"other_thread": other_thread.get_absolute_url(), "poll": other_poll.pk},
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "id": other_thread.id,
                "title": other_thread.title,
                "url": other_thread.get_absolute_url(),
            },
        )

        # other thread has two posts and an event now
        self.assertEqual(other_thread.post_set.count(), 3)

        # first thread is gone
        with self.assertRaises(Thread.DoesNotExist):
            Thread.objects.get(pk=self.thread.pk)

        # other poll and its votes are gone
        self.assertEqual(Poll.objects.filter(thread=self.thread).count(), 0)
        self.assertEqual(PollVote.objects.filter(thread=self.thread).count(), 0)

        self.assertEqual(Poll.objects.filter(thread=other_thread).count(), 1)
        self.assertEqual(PollVote.objects.filter(thread=other_thread).count(), 4)

        Poll.objects.get(pk=other_poll.pk)
        with self.assertRaises(Poll.DoesNotExist):
            Poll.objects.get(pk=poll.pk)



from unittest.mock import Mock

import pytest

from ..models import Category
from ..utils import get_categories_tree, get_category_path


@pytest.fixture
def categories_tree(root_category, default_category):
    Category(name="Category A", slug="category-a").insert_at(
        root_category, position="last-child", save=True
    )
    Category(name="Category E", slug="category-e").insert_at(
        root_category, position="last-child", save=True
    )

    category_a = Category.objects.get(slug="category-a")

    Category(name="Category B", slug="category-b").insert_at(
        category_a, position="last-child", save=True
    )

    category_b = Category.objects.get(slug="category-b")

    category_c = Category(name="Subcategory C", slug="subcategory-c").insert_at(
        category_b, position="last-child", save=True
    )
    category_d = Category(name="Subcategory D", slug="subcategory-d").insert_at(
        category_b, position="last-child", save=True
    )

    category_e = Category.objects.get(slug="category-e")
    Category(name="Subcategory F", slug="subcategory-f").insert_at(
        category_e, position="last-child", save=True
    )

    return {
        "root": root_category,
        "first": default_category,
        "a": Category.objects.get(slug="category-a"),
        "b": Category.objects.get(slug="category-b"),
        "c": Category.objects.get(slug="subcategory-c"),
        "d": Category.objects.get(slug="subcategory-d"),
        "e": Category.objects.get(slug="category-e"),
        "f": Category.objects.get(slug="subcategory-f"),
    }


@pytest.fixture
def full_access_user_acl(categories_tree, user_acl):
    user_acl = user_acl.copy()
    categories_acl = {"categories": {}, "visible_categories": []}
    for category in Category.objects.all_categories():
        categories_acl["visible_categories"].append(category.id)
        categories_acl["categories"][category.id] = {"can_see": 1, "can_browse": 1}
    user_acl.update(categories_acl)
    return user_acl


@pytest.fixture
def request_mock(user, full_access_user_acl, dynamic_settings):
    return Mock(settings=dynamic_settings, user=user, user_acl=full_access_user_acl)


def test_tree_getter_defaults_to_returning_top_level_categories(
    request_mock, categories_tree
):
    assert get_categories_tree(request_mock) == [
        categories_tree["first"],
        categories_tree["a"],
        categories_tree["e"],
    ]


def test_tree_getter_returns_category_subtree(request_mock, categories_tree):
    assert get_categories_tree(request_mock, categories_tree["a"]) == [
        categories_tree["b"]
    ]


def test_tree_getter_returns_empty_list_for_leaf_category(
    request_mock, categories_tree
):
    assert get_categories_tree(request_mock, categories_tree["f"]) == []


def test_path_getter_returns_path_to_category(request_mock, categories_tree):
    for node in get_categories_tree(request_mock):
        parent_nodes = len(get_category_path(node))
        assert parent_nodes == node.level



from django import forms
from django.utils.translation import gettext_lazy as _

from .base import OAuthProviderForm


class TwitterForm(OAuthProviderForm):
    key = forms.CharField(label=_("Consumer API key"), required=False)
    secret = forms.CharField(label=_("Consumer API secret key"), required=False)



from django.contrib.auth import get_user_model
from django.core.exceptions import PermissionDenied
from django.shortcuts import get_object_or_404, render
from django.urls import reverse
from django.utils.translation import gettext as _

from ...core.exceptions import Banned
from ..bans import get_user_ban
from ..decorators import deny_banned_ips
from ..tokens import is_password_change_token_valid


@deny_banned_ips
def request_reset(request):
    if request.settings.enable_oauth2_client:
        raise PermissionDenied(
            _("Please use %(provider)s to reset your password.")
            % {"provider": request.settings.oauth2_provider}
        )

    request.frontend_context.update(
        {"SEND_PASSWORD_RESET_API": reverse("misago:api:send-password-form")}
    )
    return render(request, "misago/forgottenpassword/request.html")


class ResetError(Exception):
    pass


@deny_banned_ips
def reset_password_form(request, pk, token):
    if request.settings.enable_oauth2_client:
        raise PermissionDenied(
            _("Please use %(provider)s to reset your password.")
            % {"provider": request.settings.oauth2_provider}
        )

    requesting_user = get_object_or_404(get_user_model(), pk=pk)

    try:
        if request.user.is_authenticated and request.user.id != requesting_user.id:
            message = _(
                "%(user)s, your link has expired. "
                "Please request new link and try again."
            )
            raise ResetError(message % {"user": requesting_user.username})

        if not is_password_change_token_valid(requesting_user, token):
            message = _(
                "%(user)s, your link is invalid. Please try again or request new link."
            )
            raise ResetError(message % {"user": requesting_user.username})

        ban = get_user_ban(requesting_user, request.cache_versions)
        if ban:
            raise Banned(ban)
    except ResetError as e:
        return render(
            request,
            "misago/forgottenpassword/error.html",
            {"message": e.args[0]},
            status=400,
        )

    api_url = reverse(
        "misago:api:change-forgotten-password", kwargs={"pk": pk, "token": token}
    )

    request.frontend_context["CHANGE_PASSWORD_API"] = api_url
    return render(request, "misago/forgottenpassword/form.html")



import json

from django.urls import reverse

from .. import test
from ...categories.models import Category
from ...users.test import AuthenticatedUserTestCase


class ThreadPollApiTestCase(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()

        self.category = Category.objects.get(slug="first-category")
        self.thread = test.post_thread(self.category, poster=self.user)

        self.api_link = reverse(
            "misago:api:thread-poll-list", kwargs={"thread_pk": self.thread.pk}
        )

    def post(self, url, data=None):
        return self.client.post(
            url, json.dumps(data or {}), content_type="application/json"
        )

    def put(self, url, data=None):
        return self.client.put(
            url, json.dumps(data or {}), content_type="application/json"
        )

    def mock_poll(self):
        self.poll = self.thread.poll = test.post_poll(self.thread, self.user)

        self.api_link = reverse(
            "misago:api:thread-poll-detail",
            kwargs={"thread_pk": self.thread.pk, "pk": self.poll.pk},
        )



from unittest.mock import Mock, patch

import pytest
from django.core.exceptions import ValidationError

from ..exceptions import OAuth2UserDataValidationError
from ..validation import validate_user_data


def test_new_user_valid_data_is_validated(db, dynamic_settings):
    valid_data = validate_user_data(
        Mock(settings=dynamic_settings),
        None,
        {
            "id": "1234",
            "name": "Test User",
            "email": "user@example.com",
            "avatar": None,
        },
        {},
    )

    assert valid_data == {
        "id": "1234",
        "name": "Test_User",
        "email": "user@example.com",
        "avatar": None,
    }


def test_existing_user_valid_data_is_validated(user, dynamic_settings):
    valid_data = validate_user_data(
        Mock(settings=dynamic_settings),
        user,
        {
            "id": "1234",
            "name": user.username,
            "email": user.email,
            "avatar": None,
        },
        {},
    )

    assert valid_data == {
        "id": "1234",
        "name": user.username,
        "email": user.email,
        "avatar": None,
    }


def user_noop_filter(*args):
    return None


def test_error_was_raised_for_user_data_with_without_name(db, dynamic_settings):
    with pytest.raises(OAuth2UserDataValidationError) as excinfo:
        # Custom filters disable build in filters
        with patch(
            "misago.oauth2.validation.oauth2_user_data_filters",
            [user_noop_filter],
        ):
            validate_user_data(
                Mock(settings=dynamic_settings),
                None,
                {
                    "id": "1234",
                    "name": "",
                    "email": "user@example.com",
                    "avatar": None,
                },
                {},
            )

    assert excinfo.value.error_list == [
        "Username can only contain latin alphabet letters and digits."
    ]


def test_error_was_raised_for_user_data_with_invalid_name(db, dynamic_settings):
    with pytest.raises(OAuth2UserDataValidationError) as excinfo:
        # Custom filters disable build in filters
        with patch(
            "misago.oauth2.validation.oauth2_user_data_filters",
            [user_noop_filter],
        ):
            validate_user_data(
                Mock(settings=dynamic_settings),
                None,
                {
                    "id": "1234",
                    "name": "Invalid!",
                    "email": "user@example.com",
                    "avatar": None,
                },
                {},
            )

    assert excinfo.value.error_list == [
        "Username can only contain latin alphabet letters and digits."
    ]


def test_error_was_raised_for_user_data_with_too_long_name(db, dynamic_settings):
    with pytest.raises(OAuth2UserDataValidationError) as excinfo:
        # Custom filters disable build in filters
        with patch(
            "misago.oauth2.validation.oauth2_user_data_filters",
            [user_noop_filter],
        ):
            validate_user_data(
                Mock(settings=dynamic_settings),
                None,
                {
                    "id": "1234",
                    "name": "UserName" * 100,
                    "email": "user@example.com",
                    "avatar": None,
                },
                {},
            )

    assert excinfo.value.error_list == [
        "Username cannot be longer than 200 characters."
    ]


def test_error_was_raised_for_user_data_without_email(db, dynamic_settings):
    with pytest.raises(OAuth2UserDataValidationError) as excinfo:
        validate_user_data(
            Mock(settings=dynamic_settings),
            None,
            {
                "id": "1234",
                "name": "Test User",
                "email": "",
                "avatar": None,
            },
            {},
        )

    assert excinfo.value.error_list == ["Enter a valid email address."]


def test_error_was_raised_for_user_data_with_invalid_email(db, dynamic_settings):
    with pytest.raises(OAuth2UserDataValidationError) as excinfo:
        validate_user_data(
            Mock(settings=dynamic_settings),
            None,
            {
                "id": "1234",
                "name": "Test User",
                "email": "userexample.com",
                "avatar": None,
            },
            {},
        )

    assert excinfo.value.error_list == ["Enter a valid email address."]


def custom_oauth2_validator(request, user, user_data, raw_data):
    if "bad" in user_data["name"].lower():
        raise ValidationError("Custom validation error!")


def test_custom_oauth2_validator_passes_valid_data(db, dynamic_settings):
    user_data = {
        "id": "1234",
        "name": "UserName",
        "email": "user@example.com",
        "avatar": None,
    }

    with patch(
        "misago.oauth2.validation.oauth2_validators",
        [custom_oauth2_validator],
    ):
        assert (
            validate_user_data(
                Mock(settings=dynamic_settings),
                None,
                user_data,
                {},
            )
            == user_data
        )


def test_custom_oauth2_validator_raises_error_for_invalid_data(db, dynamic_settings):
    with pytest.raises(OAuth2UserDataValidationError) as excinfo:
        with patch(
            "misago.oauth2.validation.oauth2_validators",
            [custom_oauth2_validator],
        ):
            validate_user_data(
                Mock(settings=dynamic_settings),
                None,
                {
                    "id": "1234",
                    "name": "UserNameBad",
                    "email": "user@example.com",
                    "avatar": None,
                },
                {},
            )

    assert excinfo.value.error_list == ["Custom validation error!"]



import os
import sys
from typing import Dict, List, Optional, Tuple


def load_plugin_list_if_exists(path: str) -> Optional[List[str]]:
    if not os.path.exists(path):
        return None

    plugins = []
    for plugin in load_plugin_list(path):
        if "@" in plugin:
            plugin, path = plugin.split("@", 1)
            sys.path.append(path)
        plugins.append(plugin)

    return plugins


def load_plugin_list(path: str) -> List[str]:
    with open(path, "r") as f:
        data = f.read()
        return parse_plugins_list(data)


def parse_plugins_list(data: str) -> List[str]:
    plugins: List[str] = []
    modules: Dict[str, Tuple[int, str]] = {}

    for line, entry in enumerate(data.splitlines()):
        plugin = entry.strip()

        if "#" in plugin:
            comment_start = plugin.find("#")
            plugin = plugin[:comment_start].strip()
        if not plugin:
            continue

        if "@" in plugin:
            module, path = map(lambda x: x.strip(), plugin.split("@", 1))
            plugin = f"{module}@{path}"
            validate_local_plugin(line, module, path)
        else:
            module = plugin
        if module in modules:
            first_line, first_entry = modules[module]
            raise ValueError(
                f"plugin '{module}' is listed more than once: "
                f"at line {first_line} ('{first_entry}') and at {line} ('{entry}')"
            )
        modules[module] = line, entry

        plugins.append(plugin)

    return plugins


def validate_local_plugin(line: int, module: str, path: str):
    if not module:
        raise ValueError(f"local plugin entry at line {line} is missing a module name")
    if not path:
        raise ValueError(f"local plugin entry at line {line} is missing a path")



from .providers import providers


def add_acl_to_obj(user_acl, obj):
    """add valid ACL to obj (iterable of objects or single object)"""
    if hasattr(obj, "__iter__"):
        for item in obj:
            _add_acl_to_obj(user_acl, item)
    else:
        _add_acl_to_obj(user_acl, obj)


def _add_acl_to_obj(user_acl, obj):
    """add valid ACL to single obj, helper for add_acl function"""
    obj.acl = {}

    for annotator in providers.get_obj_type_annotators(obj):
        annotator(user_acl, obj)



from django.db import migrations
from django.utils.translation import gettext

from ...core.utils import slugify

_ = lambda s: s


def create_default_ranks(apps, schema_editor):
    Rank = apps.get_model("misago_users", "Rank")

    team = Rank.objects.create(
        name=gettext("Forum team"),
        slug=slugify(gettext("Forum team")),
        title=gettext("Team"),
        css_class="primary",
        is_tab=True,
        order=0,
    )

    member = Rank.objects.create(
        name=gettext("Members"),
        slug=slugify(gettext("Members")),
        is_default=True,
        order=1,
    )

    Role = apps.get_model("misago_acl", "Role")

    team.roles.add(Role.objects.get(name=_("Moderator")))
    team.roles.add(Role.objects.get(name=_("Private threads")))
    team.roles.add(Role.objects.get(name=_("Private threads moderator")))
    team.roles.add(Role.objects.get(name=_("Deleting users")))

    member.roles.add(Role.objects.get(name=_("Private threads")))


class Migration(migrations.Migration):

    dependencies = [
        ("misago_users", "0003_bans_version_tracker"),
        ("misago_acl", "0003_default_roles"),
    ]

    operations = [migrations.RunPython(create_default_ranks)]



from io import StringIO

from django.contrib.auth import get_user_model
from django.core.management import call_command

from ..management.commands import createfakeusers

User = get_user_model()


def test_management_command_creates_fake_users(db):
    call_command(createfakeusers.Command(), stdout=StringIO())
    assert User.objects.exists()



from django.urls import path
from django.utils.translation import gettext_lazy as _

from .views import index
from .views import (
    ChangeAnalyticsSettingsView,
    ChangeCaptchaSettingsView,
    ChangeGeneralSettingsView,
    ChangeOAuth2SettingsView,
    ChangeThreadsSettingsView,
    ChangeUsersSettingsView,
)


class MisagoAdminExtension:
    def register_urlpatterns(self, urlpatterns):
        urlpatterns.namespace("settings/", "settings")

        urlpatterns.patterns("settings", path("", index, name="index"))

        urlpatterns.single_pattern(
            "analytics/",
            "analytics",
            "settings",
            ChangeAnalyticsSettingsView.as_view(),
        )
        urlpatterns.single_pattern(
            "captcha/", "captcha", "settings", ChangeCaptchaSettingsView.as_view()
        )
        urlpatterns.single_pattern(
            "general/", "general", "settings", ChangeGeneralSettingsView.as_view()
        )
        urlpatterns.single_pattern(
            "oauth2/", "oauth2", "settings", ChangeOAuth2SettingsView.as_view()
        )
        urlpatterns.single_pattern(
            "threads/", "threads", "settings", ChangeThreadsSettingsView.as_view()
        )
        urlpatterns.single_pattern(
            "users/", "users", "settings", ChangeUsersSettingsView.as_view()
        )

    def register_navigation_nodes(self, site):
        site.add_node(
            name=_("Settings"),
            icon="fa fa-cog",
            after="themes:index",
            namespace="settings",
        )

        site.add_node(
            name=_("General"),
            description=_("Change forum details like name, description or footer."),
            parent="settings",
            namespace="general",
        )
        site.add_node(
            name=_("Users"),
            description=_(
                "Customize user accounts default behavior and features availability."
            ),
            parent="settings",
            namespace="users",
            after="general:index",
        )
        site.add_node(
            name=_("Captcha"),
            description=_(
                "Setup protection against automatic registrations on the site."
            ),
            parent="settings",
            namespace="captcha",
            after="users:index",
        )
        site.add_node(
            name=_("Analytics"),
            description=_("Enable Google Analytics or setup Google Site Verification."),
            parent="settings",
            namespace="analytics",
            after="captcha:index",
        )
        site.add_node(
            name=_("Threads"),
            description=_("Threads, posts, polls and attachments options."),
            parent="settings",
            namespace="threads",
            after="analytics:index",
        )
        site.add_node(
            name=_("OAuth2"),
            description=_(
                "Enable OAuth2 client and connect your site to existing auth provider."
            ),
            parent="settings",
            namespace="oauth2",
            after="threads:index",
        )



from django import forms
from django.utils.translation import gettext_lazy as _

from ...admin.forms import YesNoSwitch
from ..models import MenuItem
from ..cache import clear_menus_cache


class MenuItemForm(forms.ModelForm):
    title = forms.CharField(label=_("Title"))
    url = forms.URLField(
        label=_("URL"), help_text=_("URL where this item will point to.")
    )
    menu = forms.ChoiceField(
        label=_("Menu"),
        choices=MenuItem.MENU_CHOICES,
        help_text=_("Menu in which this item will be displayed."),
    )
    css_class = forms.CharField(
        label=_("CSS class"),
        help_text=_('If you want to set custom value for link\'s "class".'),
        required=False,
    )
    target_blank = YesNoSwitch(
        label=_("Open this link in new window"),
        help_text=_(
            'Enabling this option will result in the target="_blank" attribute being '
            "added to this link's HTML element."
        ),
        required=False,
    )
    rel = forms.CharField(
        label=_("Rel attribute"),
        help_text=_(
            'Optional "rel" attribute that this item will use (ex. "nofollow").'
        ),
        required=False,
    )

    class Meta:
        model = MenuItem
        fields = ["title", "url", "menu", "css_class", "target_blank", "rel"]

    def save(self):
        item = super().save()
        clear_menus_cache()
        return item



from django.urls import reverse
from django.utils.translation import gettext as _
from django.utils.translation import ngettext
from rest_framework import serializers

__all__ = ["NewVoteSerializer", "PollVoteSerializer"]


class NewVoteSerializer(serializers.Serializer):
    choices = serializers.ListField(child=serializers.CharField())

    def validate_choices(self, data):
        if len(data) > self.context["allowed_choices"]:
            message = ngettext(
                "This poll disallows voting for more than %(choices)s choice.",
                "This poll disallows voting for more than %(choices)s choices.",
                self.context["allowed_choices"],
            )
            raise serializers.ValidationError(
                message % {"choices": self.context["allowed_choices"]}
            )

        valid_choices = [c["hash"] for c in self.context["choices"]]
        clean_choices = []

        for choice in data:
            if choice in valid_choices and choice not in clean_choices:
                clean_choices.append(choice)

        if len(clean_choices) != len(data):
            raise serializers.ValidationError(
                _("One or more of poll choices were invalid.")
            )

        if not clean_choices:
            raise serializers.ValidationError(_("You have to make a choice."))

        return clean_choices


class PollVoteSerializer(serializers.Serializer):
    voted_on = serializers.DateTimeField()
    username = serializers.SerializerMethodField()

    url = serializers.SerializerMethodField()

    class Meta:
        fields = ["voted_on", "username", "url"]

    def get_username(self, obj):
        return obj["voter_name"]

    def get_url(self, obj):
        if obj["voter_id"]:
            return reverse(
                "misago:user", kwargs={"pk": obj["voter_id"], "slug": obj["voter_slug"]}
            )



from django.contrib import messages
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from ...admin.views import generic
from ..models import Agreement
from .forms import AgreementForm, FilterAgreementsForm
from .utils import disable_agreement, set_agreement_as_active


class AgreementAdmin(generic.AdminBaseMixin):
    root_link = "misago:admin:settings:agreements:index"
    model = Agreement
    form_class = AgreementForm
    templates_dir = "misago/admin/agreements"
    message_404 = _("Requested agreement does not exist.")

    def handle_form(self, form, request, target):
        form.save()

        if self.message_submit:
            messages.success(
                request, self.message_submit % {"title": target.get_final_title()}
            )


class AgreementsList(AgreementAdmin, generic.ListView):
    items_per_page = 30
    ordering = [("-id", _("From newest")), ("id", _("From oldest"))]
    filter_form = FilterAgreementsForm
    selection_label = _("With agreements: 0")
    empty_selection_label = _("Select agreements")
    mass_actions = [
        {
            "action": "delete",
            "name": _("Delete agreements"),
            "confirmation": _("Are you sure you want to delete those agreements?"),
        }
    ]

    def get_queryset(self):
        qs = super().get_queryset()
        return qs.select_related()

    def action_delete(self, request, items):
        items.delete()
        Agreement.objects.invalidate_cache()
        messages.success(request, _("Selected agreements have been deleted."))


class NewAgreement(AgreementAdmin, generic.ModelFormView):
    message_submit = _('New agreement "%(title)s" has been saved.')

    def handle_form(self, form, request, target):
        super().handle_form(form, request, target)

        form.instance.set_created_by(request.user)
        form.instance.save()
        Agreement.objects.invalidate_cache()


class EditAgreement(AgreementAdmin, generic.ModelFormView):
    message_submit = _('Agreement "%(title)s" has been edited.')

    def handle_form(self, form, request, target):
        super().handle_form(form, request, target)

        form.instance.last_modified_on = timezone.now()
        form.instance.set_last_modified_by(request.user)
        form.instance.save()
        Agreement.objects.invalidate_cache()


class DeleteAgreement(AgreementAdmin, generic.ButtonView):
    def button_action(self, request, target):
        target.delete()
        Agreement.objects.invalidate_cache()
        message = _('Agreement "%(title)s" has been deleted.')
        messages.success(request, message % {"title": target.get_final_title()})


class SetAgreementAsActive(AgreementAdmin, generic.ButtonView):
    def button_action(self, request, target):
        set_agreement_as_active(target, commit=True)

        message = _('Agreement "%(title)s" has been set as active for type "%(type)s".')
        targets_names = {
            "title": target.get_final_title(),
            "type": target.get_type_display(),
        }
        messages.success(request, message % targets_names)


class DisableAgreement(AgreementAdmin, generic.ButtonView):
    def button_action(self, request, target):
        disable_agreement(target, commit=True)

        message = _('Agreement "%(title)s" has been disabled.') % {
            "title": target.get_final_title()
        }
        messages.success(request, message)



# Generated by Django 1.11.1 on 2017-05-21 17:52
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("misago_threads", "0005_index_search_document")]

    operations = []



import json

from django.urls import reverse

from .. import test
from ...categories.models import Category
from ...conf.test import override_dynamic_settings
from ...users.test import AuthenticatedUserTestCase
from ..models import Post, Thread
from ..test import patch_category_acl


class ThreadPostBulkPatchApiTestCase(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()

        self.category = Category.objects.get(slug="first-category")
        self.thread = test.post_thread(category=self.category)
        self.posts = [
            test.reply_thread(self.thread, poster=self.user),
            test.reply_thread(self.thread),
            test.reply_thread(self.thread, poster=self.user),
        ]

        self.ids = [p.id for p in self.posts]

        self.api_link = reverse(
            "misago:api:thread-post-list", kwargs={"thread_pk": self.thread.pk}
        )

    def patch(self, api_link, ops):
        return self.client.patch(
            api_link, json.dumps(ops), content_type="application/json"
        )


class BulkPatchSerializerTests(ThreadPostBulkPatchApiTestCase):
    def test_invalid_input_type(self):
        """api rejects invalid input type"""
        response = self.patch(self.api_link, [1, 2, 3])

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "non_field_errors": [
                    "Invalid data. Expected a dictionary, but got list."
                ]
            },
        )

    def test_missing_input_keys(self):
        """api rejects input with missing keys"""
        response = self.patch(self.api_link, {})

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"ids": ["This field is required."], "ops": ["This field is required."]},
        )

    def test_empty_input_keys(self):
        """api rejects input with empty keys"""
        response = self.patch(self.api_link, {"ids": [], "ops": []})

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "ids": ["Ensure this field has at least 1 elements."],
                "ops": ["Ensure this field has at least 1 elements."],
            },
        )

    def test_invalid_input_keys(self):
        """api rejects input with invalid keys"""
        response = self.patch(self.api_link, {"ids": ["a"], "ops": [1]})

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "ids": {"0": ["A valid integer is required."]},
                "ops": {"0": ['Expected a dictionary of items but got type "int".']},
            },
        )

    def test_too_small_id(self):
        """api rejects input with implausiple id"""
        response = self.patch(self.api_link, {"ids": [0], "ops": [{}]})

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"ids": {"0": ["Ensure this value is greater than or equal to 1."]}},
        )

    @override_dynamic_settings(posts_per_page=4, posts_per_page_orphans=3)
    def test_too_large_input(self):
        """api rejects too large input"""
        response = self.patch(
            self.api_link,
            {"ids": [i + 1 for i in range(8)], "ops": [{} for i in range(200)]},
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "ids": ["No more than 7 posts can be updated at a single time."],
                "ops": ["Ensure this field has no more than 10 elements."],
            },
        )

    def test_posts_not_found(self):
        """api fails to find posts"""
        posts = [
            test.reply_thread(self.thread, is_hidden=True),
            test.reply_thread(self.thread, is_unapproved=True),
        ]

        response = self.patch(
            self.api_link, {"ids": [p.id for p in posts], "ops": [{}]}
        )

        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "One or more posts to update could not be found."},
        )

    def test_ops_invalid(self):
        """api validates descriptions"""
        response = self.patch(self.api_link, {"ids": self.ids[:1], "ops": [{}]})

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), [{"id": self.ids[0], "detail": ["undefined op"]}]
        )

    def test_anonymous_user(self):
        """anonymous users can't use bulk actions"""
        self.logout_user()

        response = self.patch(self.api_link, {"ids": self.ids[:1], "ops": [{}]})
        self.assertEqual(response.status_code, 403)

    def test_events(self):
        """cant use bulk actions for events"""
        for post in self.posts:
            post.is_event = True
            post.save()

        response = self.patch(self.api_link, {"ids": self.ids, "ops": [{}]})

        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "One or more posts to update could not be found."},
        )


class PostsAddAclApiTests(ThreadPostBulkPatchApiTestCase):
    def test_add_acl_true(self):
        """api adds posts acls to response"""
        response = self.patch(
            self.api_link,
            {"ids": self.ids, "ops": [{"op": "add", "path": "acl", "value": True}]},
        )
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        for i, post in enumerate(self.posts):
            self.assertEqual(response_json[i]["id"], post.id)
            self.assertTrue(response_json[i]["acl"])


class BulkPostProtectApiTests(ThreadPostBulkPatchApiTestCase):
    @patch_category_acl({"can_protect_posts": True, "can_edit_posts": 2})
    def test_protect_post(self):
        """api makes it possible to protect posts"""
        response = self.patch(
            self.api_link,
            {
                "ids": self.ids,
                "ops": [{"op": "replace", "path": "is-protected", "value": True}],
            },
        )
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        for i, post in enumerate(self.posts):
            self.assertEqual(response_json[i]["id"], post.id)
            self.assertTrue(response_json[i]["is_protected"])

        for post in Post.objects.filter(id__in=self.ids):
            self.assertTrue(post.is_protected)

    @patch_category_acl({"can_protect_posts": False})
    def test_protect_post_no_permission(self):
        """api validates permission to protect posts and returns errors"""
        response = self.patch(
            self.api_link,
            {
                "ids": self.ids,
                "ops": [{"op": "replace", "path": "is-protected", "value": True}],
            },
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        for i, post in enumerate(self.posts):
            self.assertEqual(response_json[i]["id"], post.id)
            self.assertEqual(
                response_json[i]["detail"],
                ["You can't protect posts in this category."],
            )

        for post in Post.objects.filter(id__in=self.ids):
            self.assertFalse(post.is_protected)


class BulkPostsApproveApiTests(ThreadPostBulkPatchApiTestCase):
    @patch_category_acl({"can_approve_content": True})
    def test_approve_post(self):
        """api resyncs thread and categories on posts approval"""
        for post in self.posts:
            post.is_unapproved = True
            post.save()

        self.thread.synchronize()
        self.thread.save()

        self.assertNotIn(self.thread.last_post_id, self.ids)

        response = self.patch(
            self.api_link,
            {
                "ids": self.ids,
                "ops": [{"op": "replace", "path": "is-unapproved", "value": False}],
            },
        )
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        for i, post in enumerate(self.posts):
            self.assertEqual(response_json[i]["id"], post.id)
            self.assertFalse(response_json[i]["is_unapproved"])

        for post in Post.objects.filter(id__in=self.ids):
            self.assertFalse(post.is_unapproved)

        thread = Thread.objects.get(pk=self.thread.pk)
        self.assertIn(thread.last_post_id, self.ids)

        category = Category.objects.get(pk=self.category.pk)
        self.assertEqual(category.posts, 4)



import django.db.models.deletion
import mptt.fields
from django.conf import settings
from django.db import migrations, models

from ...acl.models import permissions_default


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("misago_acl", "0001_initial"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Category",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                (
                    "special_role",
                    models.CharField(max_length=255, null=True, blank=True),
                ),
                ("name", models.CharField(max_length=255)),
                ("slug", models.CharField(max_length=255)),
                ("description", models.TextField(null=True, blank=True)),
                ("is_closed", models.BooleanField(default=False)),
                ("threads", models.PositiveIntegerField(default=0)),
                ("posts", models.PositiveIntegerField(default=0)),
                (
                    "last_thread_title",
                    models.CharField(max_length=255, null=True, blank=True),
                ),
                (
                    "last_thread_slug",
                    models.CharField(max_length=255, null=True, blank=True),
                ),
                (
                    "last_poster_name",
                    models.CharField(max_length=255, null=True, blank=True),
                ),
                (
                    "last_poster_slug",
                    models.CharField(max_length=255, null=True, blank=True),
                ),
                ("last_post_on", models.DateTimeField(null=True, blank=True)),
                ("prune_started_after", models.PositiveIntegerField(default=0)),
                ("prune_replied_after", models.PositiveIntegerField(default=0)),
                ("css_class", models.CharField(max_length=255, null=True, blank=True)),
                ("lft", models.PositiveIntegerField(editable=False, db_index=True)),
                ("rght", models.PositiveIntegerField(editable=False, db_index=True)),
                ("tree_id", models.PositiveIntegerField(editable=False, db_index=True)),
                ("level", models.PositiveIntegerField(editable=False, db_index=True)),
                (
                    "archive_pruned_in",
                    models.ForeignKey(
                        related_name="pruned_archive",
                        on_delete=django.db.models.deletion.SET_NULL,
                        blank=True,
                        to="misago_categories.Category",
                        null=True,
                    ),
                ),
                (
                    "last_poster",
                    models.ForeignKey(
                        related_name="+",
                        on_delete=django.db.models.deletion.SET_NULL,
                        blank=True,
                        to=settings.AUTH_USER_MODEL,
                        null=True,
                    ),
                ),
                (
                    "parent",
                    mptt.fields.TreeForeignKey(
                        related_name="children",
                        on_delete=django.db.models.deletion.CASCADE,
                        blank=True,
                        to="misago_categories.Category",
                        null=True,
                    ),
                ),
            ],
            options={"abstract": False},
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="CategoryRole",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                (
                    "special_role",
                    models.CharField(max_length=255, null=True, blank=True),
                ),
                ("permissions", models.JSONField(default=permissions_default)),
            ],
            options={"abstract": False},
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="RoleCategoryACL",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                (
                    "category",
                    models.ForeignKey(
                        related_name="category_role_set",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="misago_categories.Category",
                    ),
                ),
                (
                    "category_role",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="misago_categories.CategoryRole",
                        to_field="id",
                    ),
                ),
                (
                    "role",
                    models.ForeignKey(
                        related_name="categories_acls",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="misago_acl.Role",
                    ),
                ),
            ],
            options={},
            bases=(models.Model,),
        ),
    ]



import re

from django import forms
from django.core.files.base import ContentFile
from django.utils.translation import gettext, gettext_lazy as _
from mptt.forms import TreeNodeChoiceField

from ...core.utils import get_file_hash
from ..models import Theme, Css
from .css import css_needs_rebuilding, create_css, get_next_css_order
from .media import create_media
from .validators import validate_css_name, validate_css_name_is_available


class ThemeChoiceField(TreeNodeChoiceField):
    level_indicator = "- - "

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("queryset", Theme.objects.all())
        kwargs.setdefault("empty_label", _("No parent"))
        kwargs.setdefault("level_indicator", self.level_indicator)
        super().__init__(*args, **kwargs)


class ThemeForm(forms.ModelForm):
    name = forms.CharField(label=_("Name"))
    parent = ThemeChoiceField(label=_("Parent"), required=False)
    version = forms.CharField(label=_("Version"), required=False)
    author = forms.CharField(label=_("Author(s)"), required=False)
    url = forms.URLField(label=_("Url"), required=False)

    class Meta:
        model = Theme
        fields = ["name", "parent", "version", "author", "url"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.limit_parent_choices()

    def limit_parent_choices(self):
        if not self.instance or not self.instance.pk:
            return

        self.fields["parent"].queryset = Theme.objects.exclude(
            tree_id=self.instance.tree_id,
            lft__gte=self.instance.lft,
            rght__lte=self.instance.rght,
        )


class ImportForm(forms.Form):
    name = forms.CharField(
        label=_("Name"),
        help_text=_("Leave this field empty to use theme name from imported file."),
        max_length=255,
        required=False,
    )
    parent = ThemeChoiceField(label=_("Parent"), required=False)
    upload = forms.FileField(
        label=_("Theme file"), help_text=_("Theme file should be a ZIP file.")
    )

    def clean_upload(self):
        data = self.cleaned_data["upload"]
        error_message = gettext("Uploaded file is not a valid ZIP file.")
        if not data.name.lower().endswith(".zip"):
            raise forms.ValidationError(error_message)
        if data.content_type not in ("application/zip", "application/octet-stream"):
            raise forms.ValidationError(error_message)
        return data


class ThemeManifest(forms.Form):
    name = forms.CharField(max_length=255)
    version = forms.CharField(max_length=255, required=False)
    author = forms.CharField(max_length=255, required=False)
    url = forms.URLField(max_length=255, required=False)


class ThemeCssUrlManifest(forms.Form):
    name = forms.CharField(max_length=255)
    url = forms.URLField(max_length=255)


def create_css_file_manifest(allowed_path):
    class ThemeCssFileManifest(forms.Form):
        name = forms.CharField(max_length=255, validators=[validate_css_name])
        path = forms.FilePathField(
            allowed_path, match=re.compile(r"\.css$", re.IGNORECASE)
        )

    return ThemeCssFileManifest


def create_media_file_manifest(allowed_path):
    class ThemeMediaFileManifest(forms.Form):
        name = forms.CharField(max_length=255)
        type = forms.CharField(max_length=255)
        path = forms.FilePathField(allowed_path)

    return ThemeMediaFileManifest


class UploadAssetsForm(forms.Form):
    allowed_content_types = []
    allowed_extensions = []

    assets = forms.FileField(
        widget=forms.ClearableFileInput(attrs={"multiple": True}),
        error_messages={"required": _("No files have been uploaded.")},
    )

    def __init__(self, *args, instance=None):
        self.instance = instance
        super().__init__(*args)

    def clean_assets(self):
        assets = []
        for asset in self.files.getlist("assets"):
            try:
                if self.allowed_content_types:
                    self.validate_asset_content_type(asset)
                if self.allowed_extensions:
                    self.validate_asset_extension(asset)
            except forms.ValidationError as e:
                self.add_error("assets", e)
            else:
                assets.append(asset)

        return assets

    def validate_asset_content_type(self, asset):
        if asset.content_type in self.allowed_content_types:
            return

        message = gettext(
            'File "%(file)s" content type "%(content_type)s" is not allowed.'
        )
        details = {"file": asset.name, "content_type": asset.content_type}

        raise forms.ValidationError(message % details)

    def validate_asset_extension(self, asset):
        filename = asset.name.lower()
        for extension in self.allowed_extensions:
            if filename.endswith(".%s" % extension):
                return

        message = gettext('File "%(file)s" extension is invalid.')
        details = {"file": asset.name}

        raise forms.ValidationError(message % details)

    def save(self):
        for asset in self.cleaned_data["assets"]:
            self.save_asset(asset)


class UploadCssForm(UploadAssetsForm):
    allowed_content_types = ["text/css"]
    allowed_extensions = ["css"]

    def save_asset(self, asset):
        create_css(self.instance, asset)


class UploadMediaForm(UploadAssetsForm):
    def save_asset(self, asset):
        create_media(self.instance, asset)


class CssEditorForm(forms.ModelForm):
    name = forms.CharField(
        label=_("Name"),
        help_text=_(
            "Should be an correct filename and include the .css extension. It will be lowercased."
        ),
        validators=[validate_css_name],
    )
    source = forms.CharField(widget=forms.Textarea(), required=False)

    class Meta:
        model = Css
        fields = ["name"]

    def clean_name(self):
        data = self.cleaned_data["name"]
        validate_css_name_is_available(self.instance, data)
        return data

    def clean(self):
        cleaned_data = super().clean()
        if not cleaned_data.get("source"):
            raise forms.ValidationError(gettext("You need to enter CSS for this file."))
        return cleaned_data

    def save(self):
        name = self.cleaned_data["name"]
        source = self.cleaned_data["source"].encode("utf-8")
        source_file = ContentFile(source, name)

        self.instance.name = name

        if self.instance.source_file:
            self.instance.source_file.delete(save=False)

        self.instance.source_file = source_file
        self.instance.source_hash = get_file_hash(source_file)
        self.instance.source_needs_building = css_needs_rebuilding(source_file)
        self.instance.size = len(source)

        if not self.instance.pk:
            self.instance.order = get_next_css_order(self.instance.theme)

        self.instance.save()
        return self.instance


class CssLinkForm(forms.ModelForm):
    name = forms.CharField(
        label=_("Link name"),
        help_text=_('Can be descriptive (e.g. "roboto from fonts.google.com").'),
    )
    url = forms.URLField(label=_("Remote CSS URL"))

    class Meta:
        model = Css
        fields = ["name", "url"]

    def clean_name(self):
        data = self.cleaned_data["name"]
        validate_css_name_is_available(self.instance, data)
        return data

    def save(self):
        if not self.instance.pk:
            self.instance.order = get_next_css_order(self.instance.theme)
            self.instance.save()
        else:
            self.instance.save(update_fields=["name", "url", "modified_on"])

        return self.instance



from ..parser import parse


def test_single_line_code_markdown(request_mock, user, snapshot):
    text = '```<script>alert("!")</script>```'
    result = parse(text, request_mock, user)
    snapshot.assert_match(result["parsed_text"])


def test_multi_line_code_markdown(request_mock, user, snapshot):
    text = """
```
<script>
alert("!")
</script>
```
    """
    result = parse(text, request_mock, user)
    snapshot.assert_match(result["parsed_text"])


def test_multi_line_code_markdown_with_language(request_mock, user, snapshot):
    text = """
```javascript
<script>
alert("!")
</script>
```
    """
    result = parse(text, request_mock, user)
    snapshot.assert_match(result["parsed_text"])



import re

from django.forms import (
    CharField,
    DateTimeField,
    RadioSelect,
    TypedChoiceField,
    ValidationError,
)
from django.core.validators import RegexValidator
from django.utils.translation import gettext_lazy as _

from ..core.utils import parse_iso8601_string


class IsoDateTimeField(DateTimeField):
    input_formats = ["iso8601"]

    def prepare_value(self, value):
        try:
            return value.isoformat()
        except AttributeError:
            return value

    def to_python(self, value):
        """
        Validates that the input can be converted to a datetime. Returns a
        Python datetime.datetime object.
        """
        if value in self.empty_values:
            return None

        try:
            return parse_iso8601_string(value)
        except ValueError:
            raise ValidationError(self.error_messages["invalid"], code="invalid")


def ColorField(**kwargs):
    return CharField(
        validators=[
            RegexValidator(
                r"^#[0-9a-f]{6}$",
                flags=re.IGNORECASE,
                message=_(
                    "Value must be a 7-character string specifying an RGB color "
                    "in a hexadecimal format."
                ),
            )
        ],
        **kwargs
    )


class YesNoSwitchBase(TypedChoiceField):
    def prepare_value(self, value):
        """normalize bools to binary 1/0 so field works on them too"""
        if value in (True, "True", "true", 1, "1"):
            return 1
        return 0

    def clean(self, value):
        return self.prepare_value(value)


def YesNoSwitch(**kwargs):
    yes_label = kwargs.pop("yes_label", _("Yes"))
    no_label = kwargs.pop("no_label", _("No"))

    return YesNoSwitchBase(
        coerce=int,
        choices=[(1, yes_label), (0, no_label)],
        widget=RadioSelect(attrs={"class": "yesno-switch"}),
        **kwargs
    )



import pytest

from ..templatetags.misago_avatars import avatar


@pytest.fixture
def avatars(user):
    return {i["size"]: i["url"] for i in user.avatars}


def test_filter_returns_url_to_200_size_image(user, avatars):
    assert avatar(user) == avatars[200]


def test_filter_returns_url_to_next_largest_image_for_given_size(user, avatars):
    assert avatar(user, 250) == avatars[400]
    assert avatar(user, 150) == avatars[200]
    assert avatar(user, 50) == avatars[100]


def test_filter_returns_url_largest_image_if_requested_size_is_not_available(
    user, avatars
):
    assert avatar(user, 500) == avatars[400]



# Generated by Django 1.11.5 on 2017-10-08 01:31
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("misago_threads", "0006_redo_partial_indexes")]

    operations = [
        migrations.AlterField(
            model_name="post",
            name="posted_on",
            field=models.DateTimeField(db_index=True),
        )
    ]



from django.core.management import BaseCommand

from ....conf.shortcuts import get_dynamic_settings
from ...signals import remove_old_ips


class Command(BaseCommand):
    help = "Removes users IPs stored for longer than configured by administrator."

    def handle(self, *args, **options):
        settings = get_dynamic_settings()
        if not settings.ip_storage_time:
            self.stdout.write("Old IP removal is disabled.")
            return

        remove_old_ips.send(sender=self, ip_storage_time=settings.ip_storage_time)

        self.stdout.write(
            "IP addresses older than %s days have been removed."
            % settings.ip_storage_time
        )



from io import StringIO

from django.contrib.auth import get_user_model
from django.core.management import call_command

User = get_user_model()


def test_superuser_is_created_if_input_is_valid(db):
    out = StringIO()

    call_command(
        "createsuperuser",
        interactive=False,
        username="test",
        email="test@example.com",
        password="password",
        stdout=out,
    )

    command_output = out.getvalue().splitlines()[-1].strip()
    user = User.objects.order_by("-id")[:1][0]

    assert command_output == ("Superuser #%s has been created successfully." % user.pk)

    assert user.username == "test"
    assert user.email == "test@example.com"
    assert user.check_password("password")



import pytest
from django.urls import reverse

from ....cache.test import assert_invalidates_cache
from ....test import assert_has_error_message
from ... import THEME_CACHE
from ..css import get_next_css_order

FIRST = 0
MIDDLE = 1
LAST = 2


@pytest.fixture
def css_list(theme):
    return [
        theme.css.create(name="CSS", url="https://test.cdn/font.css", order=FIRST),
        theme.css.create(name="CSS", url="https://test.cdn/font.css", order=MIDDLE),
        theme.css.create(name="CSS", url="https://test.cdn/font.css", order=LAST),
    ]


@pytest.fixture
def move_up(admin_client):
    def move_up_client(theme, css):
        url = reverse(
            "misago:admin:themes:move-css-up", kwargs={"pk": theme.pk, "css_pk": css.pk}
        )
        return admin_client.post(url)

    return move_up_client


@pytest.fixture
def move_down(admin_client):
    def move_down_client(theme, css):
        url = reverse(
            "misago:admin:themes:move-css-down",
            kwargs={"pk": theme.pk, "css_pk": css.pk},
        )
        return admin_client.post(url)

    return move_down_client


def test_first_css_cant_be_moved_up(move_up, theme, css_list):
    first_css = css_list[FIRST]
    move_up(theme, first_css)
    first_css.refresh_from_db()
    assert first_css.order == FIRST


def test_last_css_cant_be_moved_down(move_down, theme, css_list):
    last_css = css_list[LAST]
    move_down(theme, last_css)
    last_css.refresh_from_db()
    assert last_css.order == LAST


def test_first_css_can_be_moved_down(move_down, theme, css_list):
    first_css = css_list[FIRST]
    move_down(theme, first_css)
    first_css.refresh_from_db()
    assert first_css.order == MIDDLE


def test_last_css_can_be_moved_up(move_up, theme, css_list):
    last_css = css_list[LAST]
    move_up(theme, last_css)
    last_css.refresh_from_db()
    assert last_css.order == MIDDLE


def test_middle_css_can_be_moved_down(move_down, theme, css_list):
    middle_css = css_list[MIDDLE]
    move_down(theme, middle_css)
    middle_css.refresh_from_db()
    assert middle_css.order == LAST


def test_middle_css_can_be_moved_up(move_up, theme, css_list):
    middle_css = css_list[MIDDLE]
    move_up(theme, middle_css)
    middle_css.refresh_from_db()
    assert middle_css.order == FIRST


def test_first_css_changes_order_with_middle_css_when_moved_down(
    move_down, theme, css_list
):
    move_down(theme, css_list[FIRST])
    middle_css = css_list[MIDDLE]
    middle_css.refresh_from_db()
    assert middle_css.order == FIRST


def test_last_css_changes_order_with_middle_css_when_moved_up(move_up, theme, css_list):
    move_up(theme, css_list[LAST])
    middle_css = css_list[MIDDLE]
    middle_css.refresh_from_db()
    assert middle_css.order == LAST


def test_middle_css_changes_order_with_last_css_when_moved_down(
    move_down, theme, css_list
):
    move_down(theme, css_list[MIDDLE])
    last_css = css_list[LAST]
    last_css.refresh_from_db()
    assert last_css.order == MIDDLE


def test_middle_css_changes_order_with_first_css_when_moved_up(
    move_up, theme, css_list
):
    move_up(theme, css_list[MIDDLE])
    first_css = css_list[FIRST]
    first_css.refresh_from_db()
    assert first_css.order == MIDDLE


def test_first_css_changes_order_with_last_css_when_moved_down_after_middle_deletion(
    move_down, theme, css_list
):
    css_list[MIDDLE].delete()
    move_down(theme, css_list[FIRST])
    last_css = css_list[LAST]
    last_css.refresh_from_db()
    assert last_css.order == FIRST


def test_last_css_changes_order_with_first_css_when_moved_up_after_middle_deletion(
    move_up, theme, css_list
):
    css_list[MIDDLE].delete()
    move_up(theme, css_list[LAST])
    first_css = css_list[FIRST]
    first_css.refresh_from_db()
    assert first_css.order == LAST


def test_if_css_doesnt_belong_to_theme_move_down_action_sets_error_message(
    move_down, other_theme, css_list
):
    response = move_down(other_theme, css_list[MIDDLE])
    assert_has_error_message(response)


def test_if_css_doesnt_belong_to_theme_move_up_action_sets_error_message(
    move_up, other_theme, css_list
):
    response = move_up(other_theme, css_list[MIDDLE])
    assert_has_error_message(response)


def test_if_ran_for_default_theme_move_down_action_sets_error_message(
    move_down, default_theme, css_list
):
    response = move_down(default_theme, css_list[MIDDLE])
    assert_has_error_message(response)


def test_if_ran_for_default_theme_move_up_action_sets_error_message(
    move_up, default_theme, css_list
):
    response = move_up(default_theme, css_list[MIDDLE])
    assert_has_error_message(response)


def test_if_given_nonexisting_css_id_move_down_action_sets_error_message(
    mocker, move_down, theme, css_list
):
    response = move_down(theme, mocker.Mock(pk=css_list[LAST].pk + 1))
    assert_has_error_message(response)


def test_if_given_nonexisting_css_id_move_up_action_sets_error_message(
    mocker, move_up, theme, css_list
):
    response = move_up(theme, mocker.Mock(pk=css_list[LAST].pk + 1))
    assert_has_error_message(response)


def test_if_given_nonexisting_theme_id_move_down_action_sets_error_message(
    mocker, move_down, nonexisting_theme, css_list
):
    response = move_down(nonexisting_theme, css_list[FIRST])
    assert_has_error_message(response)


def test_if_given_nonexisting_theme_id_move_up_action_sets_error_message(
    mocker, move_up, nonexisting_theme, css_list
):
    response = move_up(nonexisting_theme, css_list[LAST])
    assert_has_error_message(response)


def test_next_new_css_order_is_larger_than_largest_existing_css_order(theme):
    theme.css.create(name="CSS", url="https://test.cdn/font.css", order=4)
    assert get_next_css_order(theme) == 5


def test_moving_css_up_invalidates_theme_cache(move_up, theme, css_list):
    with assert_invalidates_cache(THEME_CACHE):
        move_up(theme, css_list[LAST])


def test_moving_css_down_invalidates_theme_cache(move_down, theme, css_list):
    with assert_invalidates_cache(THEME_CACHE):
        move_down(theme, css_list[FIRST])



from ...models import Agreement, UserAgreement
from ..utils import disable_agreement, set_agreement_as_active


def test_activating_inactive_agreement_updates_its_flag_but_doesnt_commit_to_db(db):
    agreement = Agreement.objects.create(
        type=Agreement.TYPE_PRIVACY, link="https://somewhre.com", text="Lorem ipsum"
    )

    set_agreement_as_active(agreement)
    assert agreement.is_active

    agreement.refresh_from_db()
    assert not agreement.is_active


def test_activating_inactive_agreement_with_commit_updates_its_flag_in_db(db):
    agreement = Agreement.objects.create(
        type=Agreement.TYPE_PRIVACY, link="https://somewhre.com", text="Lorem ipsum"
    )

    set_agreement_as_active(agreement, commit=True)
    assert agreement.is_active

    agreement.refresh_from_db()
    assert agreement.is_active


def test_activating_agreement_deactivates_other_active_agreement_of_same_type(db):
    old_agreement = Agreement.objects.create(
        type=Agreement.TYPE_PRIVACY,
        link="https://somewhre.com",
        text="Lorem ipsum",
        is_active=True,
    )

    new_agreement = Agreement.objects.create(
        type=Agreement.TYPE_PRIVACY, link="https://somewhre.com", text="Lorem ipsum"
    )

    set_agreement_as_active(new_agreement, commit=True)

    old_agreement.refresh_from_db()
    new_agreement.refresh_from_db()

    assert not old_agreement.is_active
    assert new_agreement.is_active


def test_activating_agreement_doesnt_deactivate_agreement_of_other_type(db):
    agreement = Agreement.objects.create(
        type=Agreement.TYPE_PRIVACY, link="https://somewhre.com", text="Lorem ipsum"
    )

    other_type_agreement = Agreement.objects.create(
        type=Agreement.TYPE_TOS,
        link="https://somewhre.com",
        text="Lorem ipsum",
        is_active=True,
    )

    set_agreement_as_active(agreement, commit=True)

    agreement.refresh_from_db()
    other_type_agreement.refresh_from_db()

    assert agreement.is_active
    assert other_type_agreement.is_active


def test_disabling_active_agreement_updates_its_flag_but_doesnt_commit_to_db(db):
    agreement = Agreement.objects.create(
        type=Agreement.TYPE_PRIVACY,
        link="https://somewhre.com",
        text="Lorem ipsum",
        is_active=True,
    )

    disable_agreement(agreement)
    assert not agreement.is_active

    agreement.refresh_from_db()
    assert agreement.is_active


def test_disabling_active_agreement_with_commit_updates_its_flag_in_db(db):
    agreement = Agreement.objects.create(
        type=Agreement.TYPE_PRIVACY,
        link="https://somewhre.com",
        text="Lorem ipsum",
        is_active=True,
    )

    disable_agreement(agreement, commit=True)
    assert not agreement.is_active

    agreement.refresh_from_db()
    assert not agreement.is_active



from django.core.exceptions import PermissionDenied
from django.utils.translation import gettext as _
from rest_framework.permissions import BasePermission

from ...core.exceptions import Banned
from ..bans import get_request_ip_ban
from ..models import Ban


class UnbannedOnly(BasePermission):
    def is_request_banned(self, request):
        ban = get_request_ip_ban(request)
        if ban:
            hydrated_ban = Ban(
                check_type=Ban.IP,
                user_message=ban["message"],
                expires_on=ban["expires_on"],
            )
            raise Banned(hydrated_ban)

    def has_permission(self, request, view):
        self.is_request_banned(request)
        return True


class UnbannedAnonOnly(UnbannedOnly):
    def has_permission(self, request, view):
        if request.user.is_authenticated:
            raise PermissionDenied(
                _("This action is not available to signed in users.")
            )

        self.is_request_banned(request)
        return True



from datetime import timedelta
from io import StringIO

import pytest
from django.core.management import call_command
from django.utils import timezone

from ...conf.test import override_dynamic_settings
from ..management.commands import removeoldips

IP_STORE_TIME = 2


@pytest.fixture
def user_with_ip(user):
    user.joined_from_ip = "127.0.0.1"
    user.save()
    return user


@pytest.fixture
def user_with_old_ip(user_with_ip):
    joined_on_past = timezone.now() - timedelta(days=IP_STORE_TIME + 1)
    user_with_ip.joined_on = joined_on_past
    user_with_ip.save()
    return user_with_ip


def test_recent_user_joined_ip_is_not_removed_by_command(user_with_ip):
    call_command(removeoldips.Command(), stdout=StringIO())
    user_with_ip.refresh_from_db()
    assert user_with_ip.joined_from_ip


@override_dynamic_settings(ip_storage_time=IP_STORE_TIME)
def test_old_user_joined_ip_is_removed_by_command(user_with_old_ip):
    call_command(removeoldips.Command(), stdout=StringIO())
    user_with_old_ip.refresh_from_db()
    assert user_with_old_ip.joined_from_ip is None


@override_dynamic_settings(ip_storage_time=None)
def test_old_user_joined_ip_is_not_removed_by_command_if_removal_is_disabled(
    user_with_old_ip,
):
    call_command(removeoldips.Command(), stdout=StringIO())
    user_with_old_ip.refresh_from_db()
    assert user_with_old_ip.joined_from_ip


@override_dynamic_settings(ip_storage_time=IP_STORE_TIME)
def test_command_displays_message_if_old_ip_removal_is_enabled(db):
    stdout = StringIO()
    call_command(removeoldips.Command(), stdout=stdout)

    command_output = stdout.getvalue().splitlines()[0].strip()
    assert command_output == "IP addresses older than 2 days have been removed."


@override_dynamic_settings(ip_storage_time=None)
def test_command_displays_message_if_old_ip_removal_is_disabled(db):
    stdout = StringIO()
    call_command(removeoldips.Command(), stdout=stdout)

    command_output = stdout.getvalue().splitlines()[0].strip()
    assert command_output == "Old IP removal is disabled."



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots[
    "test_absolute_link_to_site_is_changed_to_relative_link 1"
] = '<p>clean_links step cleans <a href="/" target="_blank">example.com</a></p>'

snapshots[
    "test_absolute_link_to_site_without_schema_is_changed_to_relative_link 1"
] = '<p>clean_links step cleans <a href="/" target="_blank">example.com</a></p>'

snapshots[
    "test_absolute_link_with_path_to_site_is_changed_to_relative_link 1"
] = '<p>clean_links step cleans <a href="/somewhere-something/" target="_blank">example.com/somewhere-something/</a></p>'

snapshots[
    "test_local_image_is_changed_to_relative_link 1"
] = '<p>clean_links step cleans <img alt="example.com/media/img.png" src="/media/img.png" /></p>'

snapshots[
    "test_parser_converts_unmarked_links_to_hrefs 1"
] = '<p>Lorem ipsum <a href="http://test.com" rel="external nofollow noopener" target="_blank">test.com</a></p>'

snapshots[
    "test_parser_skips_links_in_code_bbcode 1"
] = "<pre><code>http://test.com</code></pre>"

snapshots[
    "test_parser_skips_links_in_inline_code_bbcode 1"
] = """<p>Lorem ipsum <br />
</p><pre><code>http://test.com</code></pre><p></p>"""

snapshots[
    "test_parser_skips_links_in_inline_code_markdown 1"
] = "<p>Lorem ipsum <code>http://test.com</code></p>"



import requests
from ariadne import QueryType
from django.core.cache import cache
from django.utils.translation import gettext as _
from requests.exceptions import RequestException

from ... import __released__, __version__
from .status import Status

CACHE_KEY = "misago_admin_version_check"
CACHE_LENGTH = 3600 * 4  # 4 hours

version_check = QueryType()


@version_check.field("version")
def resolve_version(*_):
    if not __released__:
        return get_unreleased_error()

    return check_version_with_api()


def get_unreleased_error():
    return {
        "status": Status.ERROR,
        "message": _("The site is running using unreleased version of Misago."),
        "description": _(
            "Unreleased versions of Misago can lack security features and there is "
            "no supported way to upgrade them to release versions later."
        ),
    }


def check_version_with_api():
    try:
        latest_version = get_latest_version()
        return compare_versions(__version__, latest_version)
    except (RequestException, KeyError, ValueError):
        return {
            "status": Status.WARNING,
            "message": _("Failed to connect to pypi.org API. Try again later."),
            "description": _(
                "Version check feature relies on the API operated by the Python "
                "Package Index (pypi.org) API to retrieve latest Misago release "
                "version."
            ),
        }


def get_latest_version():
    data = cache.get(CACHE_KEY)
    if not data:
        data = get_latest_version_from_api()
        cache.set(CACHE_KEY, data, CACHE_LENGTH)
    return data


def get_latest_version_from_api():
    api_url = "https://pypi.org/pypi/Misago/json"
    r = requests.get(api_url)
    r.raise_for_status()
    return r.json()["info"]["version"]


def compare_versions(current, latest):
    if latest == current:
        return {
            "status": Status.SUCCESS,
            "message": _("The site is running updated version of Misago."),
            "description": _("Misago %(version)s is latest release.")
            % {"version": current},
        }

    return {
        "status": Status.ERROR,
        "message": _("The site is running outdated version of Misago."),
        "description": _(
            "The site is running Misago version %(version)s while version %(latest)s "
            "is available."
        )
        % {"version": current, "latest": latest},
    }



from django.urls import path

from ...conf import settings
from ...core.views import home_redirect

from ..views import categories

if settings.MISAGO_THREADS_ON_INDEX:
    URL_PATH = "categories/"
else:
    URL_PATH = ""

urlpatterns = [
    path(URL_PATH, categories, name="categories"),
    # fallback for after we changed index setting
    path("categories/", home_redirect),
]



from ..markup import checksums


def is_post_valid(post):
    valid_checksum = make_post_checksum(post)
    return post.checksum == valid_checksum


def make_post_checksum(post):
    post_seeds = [str(v) for v in (post.id, str(post.posted_on.date()))]
    return checksums.make_checksum(post.parsed, post_seeds)


def update_post_checksum(post):
    post.checksum = make_post_checksum(post)
    return post.checksum
